from inverse import inv
from glob import globs
from vbweights import w
from counters import counter
from assignment import mrv_variables, initialize_mrv_vars, add_var, remove_var
from collections import deque
from learned_weights import learnedWeights

if globs['split'] == 0: # splitting based on set of base relations
   from basesplit import bsplit as split
elif globs['split'] == 1: # splitting based on horn set
   from hornsplit import hsplit as split

if globs['pcheuristic'] == 0: # path consistency
   from pc import PC as pc
elif globs['pcheuristic'] == 1: # path consistency with weighted queue
   from pcw import PCw as pc

cardmult = max(w)

# backtracking consistency
def consistency(ConMatrix, neighbors, m = -1, n = -1, repairs = None,trail =  None):  

   counter['nodes']+=1 # increment visited nodes counter

   # path consistency 
   if not pc(ConMatrix,neighbors, [], m, n, repairs, trail):
      return None

   if m == -1: 
      initialize_mrv_vars(ConMatrix)
      learnedWeights.clear()
      
   counter['decisions'] += 1

   try:
      dummy, (i,j) = min([((cardmult*split[ConMatrix[i][j]-1][0] + w[ConMatrix[i][j]-1])/(float(learnedWeights[(i,j)]) if (i,j) in learnedWeights else 1),(i,j)) for i,j in mrv_variables])
   except:
      return ConMatrix

   # splitting
   d = split[ConMatrix[i][j]-1][1]
   
   # least constraining value heuristic
   dw = [(-w[a-1],a) for a in d]
   dw.sort()
   dw = [a[1] for a in dw]

   repairs = {}
   trail = deque([])

   # as long as a consistent variable-value pair in not found, search for it
   for v in dw:
       repairs[(i,j)] = ConMatrix[i][j]
       # assignment takes place
       ConMatrix[i][j] = v
       ConMatrix[j][i] = inv[v-1]
       remove_var((i,j))
       result = consistency(ConMatrix, neighbors, i, j, repairs, trail) # keep copy of the constraint matrix in case an inconsistency happens
       if (result != None):
          return result
       # revert contraint mantrix and mrv variables to previous state
       for g, h in repairs.keys():
          ConMatrix[g][h] = repairs[(g, h)] 
          ConMatrix[h][g] = inv[repairs[(g, h)]-1]
       add_var((i,j))
       while trail:
          c,(a,b) = trail.pop()
          remove_var((a,b)) if c else add_var((a,b))            

   return None # no solution found
